from uuid import uuid4
from datetime import datetime, timedelta
import pytest
from fakeredis import FakeRedis
from omnidict import repositories
from typing import Callable
from cryptography.fernet import InvalidToken


STANDARD_REPOS: list[Callable[[], repositories.KeyValueRepository]] = [
    lambda: repositories.DictRepository(expire_seconds=60),
    lambda: repositories.DirectoryRepository(expire_seconds=60),
    lambda: repositories.DbFilenameRepository(expire_seconds=60),
    lambda: repositories.RedisRepository(FakeRedis(), expire_seconds=60),
]
ENCRYPTED_REPOS: list[Callable[[], repositories.KeyValueRepository]] = [
    lambda: repositories.DictRepository(expire_seconds=60, passphrase='test'),
    lambda: repositories.DirectoryRepository(expire_seconds=60, passphrase='test'),
    lambda: repositories.DbFilenameRepository(expire_seconds=60, passphrase='test'),
    lambda: repositories.RedisRepository(FakeRedis(), expire_seconds=60, passphrase='test'),
]
DEFAULT_REPO: list[Callable[[], repositories.KeyValueRepository]] = [
    lambda: repositories.DefaultRepository(default=uuid4)
]
ALL_REPOS = STANDARD_REPOS + ENCRYPTED_REPOS + DEFAULT_REPO


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_get_not_found(repository_factory):
    repository = repository_factory()
    with pytest.raises(KeyError):
        repository['a']


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_get_not_found_by_method(repository_factory):
    repository = repository_factory()
    assert repository.get('a') is repository.default(), '`get` non existing key'


@pytest.mark.parametrize('repository_factory', ALL_REPOS)
def test_get_existing(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    assert repository['a'] == '1', '`__getitem__` existing key'


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_get_existing_by_method(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    assert repository.get('a') == '1', '`get` existing key'


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_del_not_found(repository_factory):
    repository = repository_factory()
    with pytest.raises(KeyError):
        del repository['a']


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_del_not_found_by_method(repository_factory):
    repository = repository_factory()
    assert repository.delete('a') is None, '`delete` non existing key'


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_del_existing(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    del repository['a']


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_del_existing_by_method(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    assert repository.delete('a') is None, '`delete` existing key'


@pytest.mark.parametrize('repository_factory', ALL_REPOS)
def test_set_not_found(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'


@pytest.mark.parametrize('repository_factory', ALL_REPOS)
def test_set_not_found_by_method(repository_factory):
    repository = repository_factory()
    assert repository.set('a', '1') is None, '`set` non existing key'


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_set_existing(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    with pytest.raises(KeyError):
        repository['a'] = 1


@pytest.mark.parametrize('repository_factory',  ALL_REPOS)
def test_set_existing_by_method(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    assert repository.set('a', '1') is None, '`set` existing key'


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_get_expired_key(repository_factory, freezer):
    repository = repository_factory()
    repository['a'] = '1'
    freezer.move_to(timedelta(seconds=61))
    with pytest.raises(KeyError):
        repository['a']


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_get_expired_key_by_method(repository_factory, freezer):
    repository = repository_factory()
    repository['a'] = '1'
    freezer.move_to(timedelta(seconds=61))
    assert repository.get('a') == repository.default(), 'get expired key is not default'


@pytest.mark.parametrize('repository_factory', STANDARD_REPOS + ENCRYPTED_REPOS)
def test_get_key_refreshes_expiration(repository_factory, freezer):
    repository: repositories.KeyValueRepository = repository_factory()
    start = datetime.now()
    repository['a'] = '1'
    freezer.move_to(timedelta(seconds=40))
    _ = repository.get('a')
    freezer.move_to(timedelta(seconds=40))
    val = repository.get('a')
    assert (datetime.now() - start) > timedelta(seconds=repository.expire_seconds), (
        'Not enough time elapsed'
    )
    assert val == '1', 'get does not refreshes expiration'


@pytest.mark.parametrize('repository_factory', ENCRYPTED_REPOS)
def test_encrypted_values(repository_factory):
    repository = repository_factory()
    repository['a'] = '1'
    repository.cipher = repository.build_cipher('other')
    with pytest.raises(InvalidToken):
        repository.get('a')



from unittest import mock
from unittest import TestCase
from collections import defaultdict
import pytest
from omnidict import caching
from omnidict import repositories


cache = caching.Cache(repositories.DictRepository(expire_seconds=60))
class Case(TestCase):

    call_counter: dict[str, int] = defaultdict(int)

    @staticmethod
    def f(a, b, /, c, d, *, e, f) -> int:
        Case.call_counter['f'] += 1
        return 0

    def m(self, a, /, b, *, c) -> int:
        Case.call_counter['m'] += 1
        return 0

    @staticmethod
    @cache
    def cf(a, b, /, c, d, *, e, f) -> int:
        Case.call_counter['cf'] += 1
        return 0

    @cache
    def cm(self, a, /, b, *, c) -> int:
        Case.call_counter['cm'] += 1
        return 0


case = Case()



def test_key_from_function_call():

    def tester(func):

        def inner(*args, **kwargs):
            return caching.Cache.key_from_function_call(func, *args, **kwargs)

        return inner

    key_1 = tester(case.f)(1, 2, c=3, d=4, e=5, f=6)
    case.assertEqual(key_1, "test_caching.f:?:1:::2:&:f=6:::e=5:::d=4:::c=3")
    key_2 = tester(case.f)(1, 2, 3, e=5, f=6, d=4)
    case.assertEqual(key_1, key_2, "Equivalent function calls are different")
    key_3 = tester(case.m)(1, c=3, b=2)
    case.assertEqual(key_3, 'test_caching.m:?:1:&:c=3:::b=2')


def test_funcion_call_from_key():
    key_1 = "test_caching.f:?:1:::2:&:f=6:::e=5:::d=4:::c=3"
    _, args_1, kwargs_1 = caching.Cache.function_call_from_key(key_1)
    case.assertTupleEqual(args_1, ('1', '2'))
    case.assertDictEqual(kwargs_1, {'f': '6', 'e': '5', 'd': '4', 'c': '3'})
    key_3 = 'test_caching.m:?:1:&:c=3:::b=2'
    _, args_3, kwargs_3 = caching.Cache.function_call_from_key(key_3)
    case.assertTupleEqual(args_3, ('1',))
    case.assertDictEqual(kwargs_3, {'c': '3', 'b': '2'})


@pytest.mark.parametrize('call, call_key', [
    (lambda: case.cf(1, 2, 3, 4, e=5, f=6), 'cf'),
    (lambda: case.cm(1, 2, c=3), 'cm'),
])
def test_cache_calls_function_when_not_found(call, call_key):
    prev_counter = case.call_counter[call_key]
    result = call()
    assert result == 0, 'Result differs from expected'
    assert (prev_counter + 1) == case.call_counter['cf'], 'Cached method not called'


@pytest.mark.parametrize('call_a, call_b, call_key', [
    (lambda: case.cf(1, 2, 3, 4, e=5, f=6), lambda: case.cf(1, 2, c=3, d=4, e=5, f=6), 'cf'),
    (lambda: case.cm(1, 2, c=3), lambda: case.cm(1, b=2, c=3), 'cm'),
])
def test_cache_does_not_call_function_when_found(call_a, call_b, call_key):
    result = call_a()
    prev_counter = case.call_counter[call_key]
    result = call_b()
    assert result == 0, 'Result differs from expected'
    assert prev_counter == case.call_counter[call_key], 'Cached method called'


@pytest.mark.parametrize('call_a, call_b, call_key', [
    (lambda: case.cf(1, 2, 3, 4, e=5, f=6), lambda: case.cf(1, 2, c=3, d=4, e=5, f=6, ignore_cache=True), 'cf'),
    (lambda: case.cm(1, 2, c=3), lambda: case.cm(1, b=2, c=3, ignore_cache=True), 'cm'),
])
def test_cache_calls_function_when_cache_ignored(call_a, call_b, call_key):
    result = call_a()
    prev_counter = case.call_counter[call_key]
    result = call_b()
    assert result == 0, 'Result differs from expected'
    assert (prev_counter + 1) == case.call_counter[call_key], 'Cached method not called'



"""Defines package version dynamically"""

import re
import sys
import subprocess
from pathlib import Path
from packaging.version import Version


_PATTERN = re.compile(r"(def _get_version_from_git_tags\(default: str = )'.*'(\):)")


def _get_version_from_git_tags(default: str = '1.0.0a4'):

    def set_version_as_default(version: str):
        content = Path(__file__).read_text()
        new_content = re.sub(_PATTERN, r'\1' + f"{version!r}" + r'\2', content)
        Path(__file__).write_text(new_content)

    cmd = subprocess.run(['git', 'tag'], capture_output=True)
    versions = sorted(map(Version, cmd.stdout.decode().splitlines()))
    if not versions:
        if not default:
            raise SystemError('Cannot resolve package version')
        return default
    set_version_as_default(version := str(versions[-1]))
    return version


# If you do following line, function is called everytime package is imported
# but using class approach, only when accessing `__version__` function is called

#  __version__ = _get_version_from_git_tags()


class This(sys.__class__):

    @property
    def __version__(self) -> str:
        return _get_version_from_git_tags()


sys.modules[__name__].__class__ = This



"""
Storage classes that unifies interactions with key-value stored information
"""
# Builtins
import abc
import json
import pickle
import random
import shelve
import tempfile
from hashlib import md5
from pathlib import Path
from base64 import b64encode
from datetime import datetime, timedelta
# Installed
from cryptography.fernet import Fernet
# Types
from typing import Any
from typing import Self
from typing import Generic
from typing import TypeVar
from typing import Optional
from typing import Callable


_T = TypeVar('_T')
_END_OF_TIME = datetime.utcnow() + timedelta(seconds=3600 * 24 * 365.25 * 10)
_SeedType = None | int | float | str | bytes | bytearray


class KeyValueRepository(abc.ABC, Generic[_T]):
    """Base class for all repositories

    Repositories defined a :attr:`KeyValueRepository.storage` which is the actual
    key-value storage used to memorize data.

    Repositories might define :attr:`KeyValueRepository.expire_seconds` which defines
    the number of seconds a key is preserved in the storage **after last time is was
    gotten**.

    Repositories might define :attr:`KeyValueRepository.default` which ensures that
    :class:`KeyError` is not raised when value requested through
    :meth:`KeyValueRepository.__getitem__` does not exists.
    """

    def __init__(
        self,
        storage: _T,
        /,
        expire_seconds: int = 0,
        passphrase: _SeedType = '',
        *,
        default: None | Callable[[], Any] = None
    ):
        self.storage: _T = storage
        self.expire_seconds: int = expire_seconds
        if default is None:
            self._has_default = False
            self.default = lambda: None
        else:
            self._has_default = True
            self.default = default
        if not passphrase:
            self._is_encrypted = False
            self.cipher = None
        else:
            self._is_encrypted = True
            self.cipher = self.build_cipher(passphrase)

    @property
    def has_default(self) -> bool:
        return self._has_default

    @property
    def is_encrypted(self) -> bool:
        return self._is_encrypted

    def __repr__(self) -> str:
        return f"{type(self).__name__}(storage={repr(self.storage)})"

    def __init_subclass__(cls: type[Self]) -> None:
        super().__init_subclass__()
        delitem = cls.__delitem__
        cls.__delitem__ = lambda s, k: delitem(s, cls.customize_key(k))
        getitem = cls.__getitem__
        cls.__getitem__ = lambda s, k: cls.unserialize_val(
            getitem(s, cls.customize_key(k)) if not s.is_encrypted
            else s.cipher.decrypt(getitem(s, cls.customize_key(k)))
        )
        setitem = cls.__setitem__
        cls.__setitem__ = lambda s, k, v: setitem(
            s, cls.customize_key(k),
            cls.serialize_val(v) if not s.is_encrypted
            else s.cipher.encrypt(cls.serialize_val(v))
        )

    @staticmethod
    def build_cipher(passphrase: _SeedType) -> Fernet:
        random.seed(passphrase)
        return Fernet(key=b64encode(random.randbytes(32)))

    @staticmethod
    def customize_key(key: str) -> str:
        """Converts key into a different string, like :python:`"prefix:{}".format`"""
        return key

    @staticmethod
    def serialize_val(val: Any) -> bytes:
        """Serializes any Python object into bytes"""
        return json.dumps(val).encode()

    @staticmethod
    def unserialize_val(val: bytes) -> Any:
        """De-serializes bytes into original Python object"""
        return json.loads(val.decode())

    @classmethod
    def from_existing(
        cls,
        storage: _T,
        /,
        expire_seconds: int = 0,
        passphrase: _SeedType = '',
        *,
        default: None | Callable[[], Any] = None
    ) -> Self:
        raise NotImplementedError()

    def _expire(self, key: str) -> None:
        """Marks provided ``key`` to be forgotten after provided ``seconds`` """
        raise NotImplementedError(f"{type(self).__name__} does not support expire keys")

    @abc.abstractmethod
    def __getitem__(self, key: str) -> Any:
        """Gets value associated to provided ``key``, raising ``KeyError`` if not found"""
        ...

    @abc.abstractmethod
    def __delitem__(self, key: str) -> None:
        """Deletes provided ``key`` from storage, raising ``KeyError`` if not found"""
        ...

    @abc.abstractmethod
    def __setitem__(self, key: str, value: Any) -> None:
        """Sets value associated to provided ``key``, raising ``KeyError`` if exists"""
        ...

    def delete(self, key: str) -> None:
        """Deletes provided ``key`` from storage doing nothing if not found"""
        try:
            del self[key]
        except KeyError:
            pass

    def get(self, key: str) -> Any:
        """Gets value associated to provided ``key``, returning ``None`` if not found"""
        try:
            return self[key]
        except KeyError:
            return self.default()

    def set(self, key: str, value: Any) -> None:
        """Sets ``value`` associated to provided ``key`` overwritting if key exists"""
        self.delete(key)
        self[key] = value


class DictRepository(KeyValueRepository[dict]):
    """In-Memory :class:`dict` used as storage"""

    def __init__(
        self,
        storage: Optional[dict[str, bytes]] = None,
        expire_seconds: int = 0,
        passphrase: _SeedType = '',
        *,
        default: None | Callable[[str], Any] = None
    ):
        super().__init__(storage or dict(), expire_seconds, passphrase, default=default)
        self.expire_storage: dict[str, datetime] = dict()

    def _expire(self, key: str) -> None:
        if self.expire_seconds <= 0:
            return
        self.expire_storage[key] = datetime.utcnow() + timedelta(seconds=self.expire_seconds)

    def __getitem__(self, key: str) -> Any:
        value = self.storage.get(key)
        if (value is None):
            raise KeyError(f"Key '{key}' not found")
        if (
            (self.expire_seconds > 0)
            and ((ttl := self.expire_storage.get(key)) is not None)
            and (ttl < datetime.utcnow())
        ):
            del self[key]
            raise KeyError(f"Key '{key}' has expired")
        self._expire(key)
        return value

    def __delitem__(self, key: str) -> None:
        try:
            del self.storage[key]
        except KeyError:
            raise KeyError(f"Key '{key}' not found")
        self.expire_storage.pop(key, None)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self.storage:
            raise KeyError(f"Key '{key}' already stored")
        self.storage[key] = value
        self._expire(key)


class RedisRepository(KeyValueRepository['Redis']):
    R"""Redis service used as storage

    - Keys are used as provided
    - Values are :mod:`pickle`\ .dumped and loaded, to allow arbitrary Python
      objects
    - Results might not persist when Redis restarts, depending on configuration

    """
    @staticmethod
    def serialize_val(val: Any) -> bytes | str:
        return pickle.dumps(val)

    @staticmethod
    def unserialize_val(val: Any) -> bytes | str:
        return pickle.loads(val)

    def _expire(self, key: str) -> None:
        if self.expire_seconds <= 0:
            return
        return self.storage.expire(key, time=self.expire_seconds)

    def __getitem__(self, key: str) -> Any:
        val = self.storage.get(key)
        if val is None:
            raise KeyError(f"Key '{key}' not found")
        self._expire(key)
        return val

    def __delitem__(self, key: str) -> None:
        if self.storage.get(key) is None:
            raise KeyError(f"Key '{key}' not found")
        self.storage.delete(key)

    def __setitem__(self, key: str, value: Any) -> None:
        if self.storage.get(key) is not self.default():
            raise KeyError(f"Key '{key}' already stored")
        self.storage[key] = value
        self._expire(key)


class DirectoryRepository(KeyValueRepository[Path]):
    R"""Directory used as storage

    - Keys are hashed to avoid issues with file names.
    - Values are :mod:`pickle`\ .dumped and loaded, to allow arbitrary Python
      objects
    - Results are persisted in local filesystem

    """
    def __init__(
        self,
        directory: Path | str = '',
        expire_seconds: int = 0,
        passphrase: _SeedType = '',
        *,
        default: None | Callable[[str], Any] = None
    ):
        storage = Path(directory or tempfile.mkdtemp())
        if storage.is_file():
            raise TypeError(f"DirectoryRepository initialized with file path '{storage}'")
        super().__init__(storage, expire_seconds, passphrase, default=default)
        self.storage.mkdir(exist_ok=True, parents=True)
        self.expire_storage = shelve.DbfilenameShelf(tempfile.mktemp())

    @staticmethod
    def customize_key(key: str):
        return md5(key.encode()).hexdigest()

    @staticmethod
    def serialize_val(val: Any) -> bytes | str:
        return pickle.dumps(val)

    @staticmethod
    def unserialize_val(val: Any) -> bytes | str:
        return pickle.loads(val)

    def _expire(self, key: str) -> None:
        if self.expire_seconds <= 0:
            return
        self.expire_storage[key] = datetime.utcnow() + timedelta(seconds=self.expire_seconds)

    def __getitem__(self, key: str) -> Any:
        filepath = self.storage / f"{key}.pickle"
        if not filepath.exists():
            raise KeyError(f"Key '{key}' not found")
        if (
            (self.expire_seconds > 0)
            and ((ttl := self.expire_storage.get(key)) is not None)
            and (ttl < datetime.utcnow())
        ):
            del self[key]
            raise KeyError(f"Key '{key}' has expired")
        self._expire(key)
        return filepath.read_bytes()

    def __delitem__(self, key: str) -> None:
        filepath = self.storage / f"{key}.pickle"
        if not filepath.exists():
            raise KeyError(f"Key '{key}' not found")
        filepath.unlink()

    def __setitem__(self, key: str, value: Any) -> None:
        filepath = self.storage / f"{key}.pickle"
        if filepath.exists():
            raise KeyError(f"Key '{key}' already stored")
        self._expire(key)
        filepath.write_bytes(value)


class DbFilenameRepository(KeyValueRepository[Path]):
    R"""Db filename used as storage

    - Keys are preserved
    - Values are :mod:`pickle`\ .dumped and loaded, to allow arbitrary Python
      objects
    - Results are persisted in local filesystem
    """
    def __init__(
        self,
        filepath: str | Path = '',
        expire_seconds: int = 0,
        passphrase: _SeedType = '',
        *,
        default: None | Callable[[str], Any] = None
    ):
        if Path(filepath).is_file():
            raise TypeError(f"DbFilenameRepository initialized with file path '{filepath}'")
        storage = shelve.DbfilenameShelf(filepath or tempfile.mktemp())
        super().__init__(storage, expire_seconds, passphrase, default=default)
        self.expire_storage = shelve.DbfilenameShelf(tempfile.mktemp())

    def _expire(self, key: str) -> None:
        if self.expire_seconds <= 0:
            return
        self.expire_storage[key] = datetime.utcnow() + timedelta(seconds=self.expire_seconds)

    def __getitem__(self, key: str) -> Any:
        value = self.storage.get(key)
        if (value is None):
            raise KeyError(f"Key '{key}' not found")
        if (
            (self.expire_seconds > 0)
            and ((ttl := self.expire_storage.get(key)) is not None)
            and (ttl < datetime.utcnow())
        ):
            del self[key]
            raise KeyError(f"Key '{key}' has expired")
        self._expire(key)
        return value

    def __delitem__(self, key: str) -> None:
        try:
            del self.storage[key]
        except KeyError:
            raise KeyError(f"Key '{key}' not found")
        self.expire_storage.pop(key, None)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self.storage:
            raise KeyError(f"Key '{key}' already stored")
        self.storage[key] = value
        self._expire(key)


class DefaultRepository(KeyValueRepository[dict]):
    """Repository that always returns values from ``default``"""

    def __init__(self, *, default: Callable[[str | bytes], str | bytes]):
        super().__init__(dict(), expire_seconds=0, default=default)

    @staticmethod
    def serialize_val(val: Any) -> bytes | str:
        return pickle.dumps(val)

    @staticmethod
    def unserialize_val(val: Any) -> bytes | str:
        return pickle.loads(val)

    def __getitem__(self, key: str) -> Any:
        """Keys are defined when retrieved from the first time"""
        try:
            return self.storage[key]
        except KeyError:
            pass
        self[key] = self.default(key)
        return self[key]

    def __delitem__(self, key: str) -> None:
        self.storage.pop(key, None)

    def __setitem__(self, key: str, value: Any) -> None:
        if key in self.storage:
            raise KeyError(f"Key '{key}' already stored")
        self.storage[key] = value



# Builtins
from functools import wraps
from itertools import starmap
# Local
from omnidict import repositories
# Types
from typing import Any
from typing import Generic
from typing import TypeVar
from typing import Optional
from typing import Callable
from typing import ParamSpec
from typing import Concatenate


_T = TypeVar('_T')
_R = TypeVar('_R', bound=repositories.KeyValueRepository)
_P = ParamSpec('_P')


class Cache(Generic[_R]):
    """Decorator class to cache function calls

    It can be instantiated with a method to validate results. Decorated function
    is called and if it returns a valid value, it is cached.

    You set ``repository`` attribute to define how cache should be stored. Default
    it uses in memory dictionary.

    When you use ``Cache`` instance to decorate a function, the keyword-only
    ``ignore_cache`` argument is added to its signature so you can avoid using
    cached values
    """
    fun_sep = ':?:'
    type_sep = ':&:'
    param_sep = ':::'

    def __init__(
        self,
        repository: Optional[_R] = None,
        result_validator: Optional[Callable[[Any], bool]] = None,
    ):
        if result_validator is not None:
            self.validate = result_validator
        else:
            self.validate = lambda _: True
        if repository is None:
            self.repository: _R = repositories.DictRepository()
        else:
            self.repository: _R = repository

    @classmethod
    def key_from_function_call(cls, func: Callable, *args, **kwargs) -> str:
        """Creates cache key for given function, arguments and keyword
        arguments
        """
        def all_as_kwargs(*args, **kwargs):
            in_args, in_kwargs = list(args), dict()
            npos = func.__code__.co_posonlyargcount
            nargs = func.__code__.co_argcount + func.__code__.co_kwonlyargcount
            varnames = list(func.__code__.co_varnames[:nargs][::-1])
            for i, varname in enumerate(varnames, start=1):
                # Take kwonly args from kwargs
                try:
                    in_kwargs[varname] = kwargs[varname]
                except KeyError:
                    pass
                else:
                    continue
                # Leave remaining args in kwargs
                if (nargs - len(in_kwargs)) == npos:
                    break
                # Move pos args to kwargs when possible
                try:
                    in_kwargs[varname] = in_args.pop(-1)
                except IndexError:
                    pass
            return in_args, in_kwargs

        prefix = f"{func.__module__}.{func.__name__}"
        in_args, in_kwargs = all_as_kwargs(*args, **kwargs)
        argstr = cls.param_sep.join(map(str, in_args))
        kwargstr = cls.param_sep.join(starmap("{}={}".format, in_kwargs.items()))
        return f"{prefix}{cls.fun_sep}{argstr}{cls.type_sep}{kwargstr}"

    @classmethod
    def function_call_from_key(
        cls,
        key: str
    ) -> tuple[str, tuple[str, ...], dict[str, str]]:
        """Recompose function call from cache key.

        .. warning::
            Every argument is returned as a string, this method do not cast
            argument types

        Args:
            key: The key to be decomposed

        Returns:
            Function fully qualified name, call arguments and call keyword
            arguments
        """
        prefix, argstr_and_kwargstr = key.split(cls.fun_sep)
        argstr, kwargstr = argstr_and_kwargstr.split(cls.type_sep)
        args = tuple(map(str, filter(bool, argstr.split(cls.param_sep))))
        kwargs_tuple = tuple(filter(bool, kwargstr.split(cls.param_sep)))
        kwargs = dict(map(lambda s: s.split('=', 1), kwargs_tuple))
        return prefix, args, kwargs

    def __call__(
        self,
        func: Callable[_P, _T]
    ) -> Callable[Concatenate[bool, _P], _T]:

        @wraps(func)
        def inner_func(
            *args: _P.args, ignore_cache: bool = False, **kwargs: _P.kwargs
        ) -> _T:
            try:
                cache_key = self.key_from_function_call(func, *args, **kwargs)
            except IndexError:
                raise TypeError(
                    f"Cannot create cache key for called {func} with {args=} "
                    f"and {kwargs=}"
                )
            # Retrieve key from cache
            if not ignore_cache:
                # Scan cache using None as default
                old_default = self.repository.default
                self.repository.default = lambda: None
                result = self.repository.get(cache_key)
                self.repository.default = old_default
                if result is not None:
                    return result
            # Retrieve key calling function
            result = func(*args, **kwargs)
            if self.validate(result):
                self.repository.set(cache_key, result)
            return result

        return inner_func
